
#include "stdlib.h"
#include "stdio.h"


#include "graphics.h"
#include "common.h"

#include <math.h>

#include <algorithm>
#include <vector>
#include <set>
using namespace std;


#include "myutils.h"
#include "sparse_matrix.h"
#include "my_graphics_stuff.h"
#include "solve.h"

#define MYUMF 1
#ifdef MYUMF
//#include "umfpack.h"
#endif

void read_netlist(char *filename);
void do_analytic_placement();
void drawscreen();
double compute_wire_length_hp();
void do_gordian_overlap();
void randomize_ios();
void run_fitter();


const int CHIP_DIM = 200;
char do_gordian;
float gordian_edge_weight;
char run_fitting;
char fit_sort_by_fanout;
int num_random_io_swaps;
bool use_cuda;

vector<int> net_id_to_user_name;
vector<int> net_user_name_to_id;
vector<int>	net_fanout_count;

vector<int> block_id_to_user_name;
vector<int> block_user_name_to_id;

vector<int> block_id_to_fixed_id;
vector<int> block_fixed_id_to_id;
vector<int> block_id_to_nonfixed_id;
vector<int> block_nonfixed_id_to_id;

vector<int> block_fixed_id_to_x;
vector<int> block_fixed_id_to_y;

float *block_nonfixed_id_to_x;
float *block_nonfixed_id_to_y;

float *block_nonfixed_id_to_x_actual;
float *block_nonfixed_id_to_y_actual;

sparse_matrix net_fanouts; //cols represent net ids, rows are block ids

//some functions want arrays, not vectors
template <class T> const T* vector_to_array(vector<T> &vec)
{
	//if vectors is stored sequentially, can just do this
	//if doesn't work, we can change to copy into an array
   return &(*vec.begin());
}

void drawblock(float x, float y, const char *text)
{
   //printf("x %f, y %f, t %s\n", x, y, text);
	MYASSERT (0 <= x && x <= CHIP_DIM);
	MYASSERT (0 <= y && y <= CHIP_DIM);
	drawrect(OFFSET + x * NEW_WORLD_WIDTH / CHIP_DIM, WORLD_HEIGHT - (OFFSET + y * NEW_WORLD_HEIGHT / CHIP_DIM), OFFSET + 4 + x * NEW_WORLD_WIDTH / CHIP_DIM, WORLD_HEIGHT - (OFFSET + 4 + y * NEW_WORLD_HEIGHT / CHIP_DIM));
	drawtext(OFFSET + x * NEW_WORLD_WIDTH / CHIP_DIM, WORLD_HEIGHT - (OFFSET + 15 + y * NEW_WORLD_HEIGHT / CHIP_DIM), buffer, 1000);
}

void drawwire(float x1, float y1, float x2, float y2)
{
	drawline(OFFSET + x1 * NEW_WORLD_WIDTH / CHIP_DIM, WORLD_HEIGHT - (OFFSET + y1 * NEW_WORLD_HEIGHT / CHIP_DIM), OFFSET + x2 * NEW_WORLD_WIDTH / CHIP_DIM, WORLD_HEIGHT - (OFFSET + y2 * NEW_WORLD_HEIGHT / CHIP_DIM));
}


void drawscreen (void) {

	/* redrawing routine for still pictures.  Graphics package calls  *
	* this routine to do redrawing after the user changes the window *
	* in any way.                                                    */

	clearscreen();  /* Should be first line of all drawscreens */
	setfontsize (10);
	setlinestyle (SOLID);

	setlinewidth (1);
	drawrect(OFFSET, OFFSET, WORLD_WIDTH - OFFSET, WORLD_HEIGHT - OFFSET);

	setlinewidth (5);



	//draw all blocks
	unsigned int block_id;
	int block_user_id;
	float x, y;

	for (block_id = 0; block_id < block_nonfixed_id_to_id.size(); ++ block_id)
	{
		x = block_nonfixed_id_to_x_actual[block_id];
		y = block_nonfixed_id_to_y_actual[block_id];
		block_user_id = block_id_to_user_name[block_nonfixed_id_to_id[block_id]];
		sprintf(buffer, "%d", block_user_id);
		drawblock((float)x,(float)y,buffer);
	}

	for (block_id = 0; block_id < block_fixed_id_to_id.size(); ++ block_id)
	{
		x = block_fixed_id_to_x[block_id];
		y = block_fixed_id_to_y[block_id];
		block_user_id = block_id_to_user_name[block_fixed_id_to_id[block_id]];
		sprintf(buffer, "%d", block_user_id);
		drawblock((float)x,(float)y,buffer);
	}

	setlinewidth(1);

	//go through and draw every net
	//unsigned int col, *rows, num_rows, irow1, irow2;
	//int block_id2;
	float x1,y1,x2,y2;
	for (unsigned int col = 0; col < net_fanouts.get_num_cols(); ++ col)
	{
		for (unsigned int irow1 = 0; irow1 < net_fanouts.get_num_rows(); ++ irow1)
		{
			if (net_fanouts.get_entry(irow1, col) == 0)
				continue;

			unsigned int block_id = irow1;
			if (block_id_to_nonfixed_id[block_id] != -1) //non fixed
			{
				x1 = (float)block_nonfixed_id_to_x_actual[block_id_to_nonfixed_id[block_id]];
				y1 = (float)block_nonfixed_id_to_y_actual[block_id_to_nonfixed_id[block_id]];
			}
			else //fixed
			{
				x1 = (float)block_fixed_id_to_x[block_id_to_fixed_id[block_id]];
				y1 = (float)block_fixed_id_to_y[block_id_to_fixed_id[block_id]];
			}

			for (unsigned int irow2 = 0; irow2 < net_fanouts.get_num_rows(); ++ irow2)
			{
				if (net_fanouts.get_entry(irow2, col) == 0)
					continue;

				unsigned int block_id2 = irow2;
				if (block_id_to_nonfixed_id[block_id2] != -1) //non fixed
				{
					x2 = (float)block_nonfixed_id_to_x_actual[block_id_to_nonfixed_id[block_id2]];
					y2 = (float)block_nonfixed_id_to_y_actual[block_id_to_nonfixed_id[block_id2]];
				}
				else //fixed
				{
					x2 = (float)block_fixed_id_to_x[block_id_to_fixed_id[block_id2]];
					y2 = (float)block_fixed_id_to_y[block_id_to_fixed_id[block_id2]];
				}
				drawwire(x1,y1,x2,y2);
			}
		}
	}
		/*
		rows = net_fanouts.get_row_indicies(col, &num_rows);
		for (irow1 = 0; irow1 < num_rows; ++ irow1)
		{
			block_id = rows[irow1];
			if (block_id_to_nonfixed_id[block_id] != -1) //non fixed
			{
				x1 = block_nonfixed_id_to_x_actual[block_id_to_nonfixed_id[block_id]];
				y1 = block_nonfixed_id_to_y_actual[block_id_to_nonfixed_id[block_id]];
			}
			else //fixed
			{
				x1 = block_fixed_id_to_x[block_id_to_fixed_id[block_id]];
				y1 = block_fixed_id_to_y[block_id_to_fixed_id[block_id]];
			}

			for (irow2 = 0; irow2 < num_rows; ++ irow2)
			{
				block_id2 = rows[irow2];
				if (block_id_to_nonfixed_id[block_id2] != -1) //non fixed
				{
					x2 = block_nonfixed_id_to_x_actual[block_id_to_nonfixed_id[block_id2]];
					y2 = block_nonfixed_id_to_y_actual[block_id_to_nonfixed_id[block_id2]];
				}
				else //fixed
				{
					x2 = block_fixed_id_to_x[block_id_to_fixed_id[block_id2]];
					y2 = block_fixed_id_to_y[block_id_to_fixed_id[block_id2]];
				}
				drawwire(x1,y1,x2,y2);
			}
		}
	}
	*/
}

float *current_distance_array;
int sort_by_distance_array(int x, int y)
{
	return current_distance_array[y] > current_distance_array[x];
}

void do_gordian_overlap()
{

	//add 4 blocks placed at the 4 corners

	int i;
	const int NUM_FAKE_BLOCKS = 4;
	int fake_block_fixed_ids[NUM_FAKE_BLOCKS];
	for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
	{
		fake_block_fixed_ids[i] = block_fixed_id_to_id.size();
		block_fixed_id_to_id.push_back(block_id_to_fixed_id.size());
		block_id_to_fixed_id.push_back(fake_block_fixed_ids[i]);
		block_id_to_nonfixed_id.push_back(-1);
	}

	block_fixed_id_to_x.push_back(0);
	block_fixed_id_to_y.push_back(0);

	block_fixed_id_to_x.push_back(0);
	block_fixed_id_to_y.push_back(100);

	block_fixed_id_to_x.push_back(100);
	block_fixed_id_to_y.push_back(0);

	block_fixed_id_to_x.push_back(100);
	block_fixed_id_to_y.push_back(100);

	float *distances_to_fake_blocks[NUM_FAKE_BLOCKS];
	int num_nonfixed_blocks = block_nonfixed_id_to_id.size();

	//now calculate the 1/4 closest blocks to each point
	//first calculate the distance between every nonfixed block and every fake block
	int iblock;
	for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
	{
		distances_to_fake_blocks[i] = new float[num_nonfixed_blocks];
		for (iblock = 0; iblock < num_nonfixed_blocks; ++ iblock)
		{
			distances_to_fake_blocks[i][iblock] = sqrt(
				pow(block_nonfixed_id_to_x[iblock] - block_fixed_id_to_x[fake_block_fixed_ids[i]], 2) +
				pow(block_nonfixed_id_to_y[iblock] - block_fixed_id_to_y[fake_block_fixed_ids[i]], 2)
			);
		}
	}

	//now provide list of block_nonfixed_ids for each fake block sorted by distance
	int *closest_blocks[NUM_FAKE_BLOCKS];
	for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
   {
		closest_blocks[i] = new int[num_nonfixed_blocks];
		for (iblock = 0; iblock < num_nonfixed_blocks; ++ iblock)
		{
			closest_blocks[i][iblock] = iblock;
		}
		current_distance_array = distances_to_fake_blocks[i];
		sort(&closest_blocks[i][0], &closest_blocks[i][num_nonfixed_blocks], sort_by_distance_array);
		delete[] distances_to_fake_blocks[i]; //done with distances now
	}

   /*
   for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
   {
      printf("fake block %d,%d:\n", block_fixed_id_to_x[fake_block_fixed_ids[i]], block_fixed_id_to_y[fake_block_fixed_ids[i]]);
	   for (iblock = 0; iblock < num_nonfixed_blocks; ++ iblock)
	   {
		   printf("block %d at %f,%f\n", block_id_to_user_name[block_nonfixed_id_to_id[closest_blocks[i][iblock]]], block_nonfixed_id_to_x[closest_blocks[i][iblock]], block_nonfixed_id_to_y[closest_blocks[i][iblock]]);
	   }
   }
   */


	//now figure out which quarter are closest
	//take turns - for each block, let it pick it's next closest
	bool *is_block_taken = new bool[num_nonfixed_blocks];
	for (iblock = 0; iblock < num_nonfixed_blocks; ++ iblock)
	{
		is_block_taken[iblock] = false;
	}

	
	int next_closest_block_index[NUM_FAKE_BLOCKS];
	for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
	{
		next_closest_block_index[i] = 0;
	}

#ifdef MYDEBUG
	float wl = (float)compute_wire_length_hp();
	sprintf(buffer, "HP W/L = %f", wl);
	my_update_message(buffer);
	my_event_loop(button_press, mouse_move, key_press, drawscreen);
#endif

	int original_num_nets = net_fanouts.get_num_cols();

	int num_blocks_taken = 0;
	int count_forever;
	//now take turns
	for (count_forever = 0; count_forever < INT_MAX && num_blocks_taken < num_nonfixed_blocks; ++ count_forever)
	{
		i = count_forever % NUM_FAKE_BLOCKS;
		while (is_block_taken[closest_blocks[i][next_closest_block_index[i]]])
		{
			++ next_closest_block_index[i];
		}
      //printf("Fake block at %d %d is picking user block %d\n", block_fixed_id_to_x[fake_block_fixed_ids[i]], block_fixed_id_to_y[fake_block_fixed_ids[i]], block_id_to_user_name[block_nonfixed_id_to_id[closest_blocks[i][next_closest_block_index[i]]]]);
		MYASSERT (i < num_nonfixed_blocks); //we must find one if we haven't exied the loop yet
		//add a new net between this nonfixed block and our fake block
		int new_net_id = net_fanouts.get_num_cols();
		net_fanouts.add_to_entry_expand(block_nonfixed_id_to_id[closest_blocks[i][next_closest_block_index[i]]], new_net_id, gordian_edge_weight);
		net_fanouts.add_to_entry_expand(block_fixed_id_to_id[fake_block_fixed_ids[i]], new_net_id, gordian_edge_weight);
		is_block_taken[closest_blocks[i][next_closest_block_index[i]]] = true;
		++ next_closest_block_index[i];
		++ num_blocks_taken;
	}

	MYASSERT (num_blocks_taken == num_nonfixed_blocks); //you never know

	delete[] is_block_taken;
	for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
	{
		delete[] closest_blocks[i];
	}

	net_fanouts.debug_print("Net fanouts Before Gordian change");

#ifdef MYDEBUG
	wl = (float)compute_wire_length_hp();
	sprintf(buffer, "After Gordian, before new AP - HP W/L = %f", wl);
	my_update_message(buffer);
	my_event_loop(button_press, mouse_move, key_press, drawscreen);
#endif

	do_analytic_placement();

#ifdef MYDEBUG
	wl = (float)compute_wire_length_hp();
	sprintf(buffer, "After Gordian, +after new AP - HP W/L = %f", wl);
	my_update_message(buffer);
	my_event_loop(button_press, mouse_move, key_press, drawscreen);
#endif

	//undo all modifications caused by Gordian
	for (i = 0; i < NUM_FAKE_BLOCKS; ++ i)
	{
		block_fixed_id_to_id.pop_back();
		block_id_to_fixed_id.pop_back();
		block_id_to_nonfixed_id.pop_back();
		block_fixed_id_to_x.pop_back();
		block_fixed_id_to_y.pop_back();
	}

	//clean up net fanouts
	//we added one new net for each nonfixed block

	//hacky, but easier and more efficient for our purposes
	/*
	net_fanouts.m_ap.resize(net_fanouts.m_ap.size() - num_nonfixed_blocks);
	//2 blocks for each net
	net_fanouts.m_rows.resize(net_fanouts.m_rows.size() - 2 * num_nonfixed_blocks);
	net_fanouts.m_vals.resize(net_fanouts.m_vals.size() - 2 * num_nonfixed_blocks);
	net_fanouts.m_num_rows = block_id_to_user_name.size();
	*/

	MYASSERT(net_fanouts.get_num_cols() == original_num_nets); //should be restored

	net_fanouts.debug_print("Net fanouts After Gordian change");
}


void do_analytic_placement()
{
	//need to setup Qx=c and Py=d sparse matrices

	//Q,P are num_unfixed X num_unfixed
	//C,D are num_unfixed X 1

	//calculate degree of each block in the nonfixed block matrix
	//start off at 1 cause every nonfixed block will also have an entry with itself
	vector<int>	block_nonfixed_fanout_count;
	block_nonfixed_fanout_count.resize(block_nonfixed_id_to_id.size(), 1);

	//since some blocks will have more than 1 edge connecting them,
	//we need to make sure we only count that fanout once
	struct UNIQUE_FANOUTS
	{
		set<int> myset;
		bool is_unique(int source_non_fixed_block_id, int dest_non_fixed_block_id)
		{
			//make simple key - we know how many fixed blocks there are
			int key = block_nonfixed_id_to_id.size() * source_non_fixed_block_id
					+	dest_non_fixed_block_id;
			if (myset.find(key) != myset.end())
			{
				return false; //already exists
			}
			else
			{
				myset.insert(key);
				return true;
			}
		}
	};

	UNIQUE_FANOUTS unique_fanouts;
	

	//int col, block_id1, block_id2, irow1, irow2, *rows, num_rows;

	for (unsigned int col = 0; col < net_fanouts.get_num_cols(); ++ col)
	{
		for (unsigned int irow1 = 0; irow1 < net_fanouts.get_num_rows(); ++ irow1)
		{
			if (net_fanouts.get_entry(irow1, col) == 0)
				continue;

			unsigned int block_id1 = irow1;
			if (block_id_to_nonfixed_id[block_id1] == -1)
			{
				continue; //is a fixed block
			}
			for (unsigned int irow2 = irow1 + 1; irow2 < net_fanouts.get_num_rows(); ++ irow2)
			{
				if (net_fanouts.get_entry(irow2, col) == 0)
					continue;

				unsigned int block_id2 = irow2;
				if (block_id_to_nonfixed_id[block_id2] == -1)
				{
					continue; //is a fixed block
				}
				if (unique_fanouts.is_unique(block_id1, block_id2))
				{
					++ block_nonfixed_fanout_count[block_id_to_nonfixed_id[block_id1]];
					++ block_nonfixed_fanout_count[block_id_to_nonfixed_id[block_id2]];
				}
			}
		}
		/*
		rows = net_fanouts.get_row_indicies(col, &num_rows);
		for (irow1 = 0; irow1 < num_rows; ++ irow1)
		{
			block_id1 = rows[irow1];
			if (block_id_to_nonfixed_id[block_id1] == -1)
			{
				continue; //is a fixed block
			}
			for (irow2 = irow1 + 1; irow2 < num_rows; ++ irow2)
			{
				MYASSERT (rows[irow1] != rows[irow2]);
				block_id2 = rows[irow2];
				if (block_id_to_nonfixed_id[block_id2] == -1)
				{
					continue; //is a fixed block
				}
				if (unique_fanouts.is_unique(block_id1, block_id2))
				{
					++ block_nonfixed_fanout_count[block_id_to_nonfixed_id[block_id1]];
					++ block_nonfixed_fanout_count[block_id_to_nonfixed_id[block_id2]];
				}
			}
		}
		*/
	}
	
   sparse_matrix Q;
	//SPARSE_MATRIX Q, P;
	//Q.init(block_nonfixed_id_to_id.size(),block_nonfixed_id_to_id.size(), vector_to_array(block_nonfixed_fanout_count));
	Q.init(block_nonfixed_id_to_id.size(),block_nonfixed_id_to_id.size());
	
	//P.init(block_nonfixed_id_to_id.size(),block_nonfixed_id_to_id.size(), vector_to_array(block_nonfixed_fanout_count));
	
	float *C = new float[block_nonfixed_id_to_id.size()];
	float *D = new float[block_nonfixed_id_to_id.size()];

	for (unsigned int block_id1 = 0; block_id1 < block_nonfixed_id_to_id.size(); ++ block_id1)
	{
		C[block_id1] = 0;
		D[block_id1] = 0;
	}

	if (block_nonfixed_id_to_x != NULL) //not the first call
	{
		delete[] block_nonfixed_id_to_x;
		delete[] block_nonfixed_id_to_y;
		delete[] block_nonfixed_id_to_x_actual;
		delete[] block_nonfixed_id_to_y_actual;
	}

	block_nonfixed_id_to_x = new float[block_nonfixed_id_to_id.size()];
	block_nonfixed_id_to_y = new float[block_nonfixed_id_to_id.size()];


	//now actually populate P,Q,C,D
	int nonfixed_block_id1, nonfixed_block_id2;
	int fixed_block_id1, fixed_block_id2;
	
	for (unsigned int col = 0; col < net_fanouts.get_num_cols(); ++ col)
	{
		for (unsigned int irow1 = 0; irow1 < net_fanouts.get_num_rows(); ++ irow1)
		{
			if (net_fanouts.get_entry(irow1, col) == 0)
				continue;
			unsigned int block_id1 = irow1;
			nonfixed_block_id1 = block_id_to_nonfixed_id[block_id1];
			fixed_block_id1 = block_id_to_fixed_id[block_id1];
			for (unsigned int irow2 = irow1 + 1; irow2 < net_fanouts.get_num_rows(); ++ irow2)
			{
				if (net_fanouts.get_entry(irow2, col) == 0)
					continue;

				unsigned int block_id2 = irow2;
				nonfixed_block_id2 = block_id_to_nonfixed_id[block_id2];
				fixed_block_id2 = block_id_to_fixed_id[block_id2];
				
				if (nonfixed_block_id1 != -1) //nonfixed
				{
					//Qii = sum Wik, all k
					Q.add_to_entry(nonfixed_block_id1, nonfixed_block_id1, net_fanouts.get_entry(block_id1, col));
					//P.add_to_entry(nonfixed_block_id1, nonfixed_block_id1, net_fanouts.get_entry(block_id1, col));

					if (nonfixed_block_id2 != -1) //also nonfixed
					{
						//Qij = Qji = -Wij
						Q.add_to_entry(nonfixed_block_id1, nonfixed_block_id2, -1 * net_fanouts.get_entry(block_id1, col));
						Q.add_to_entry(nonfixed_block_id2, nonfixed_block_id1, -1 * net_fanouts.get_entry(block_id1, col));
						//P.add_to_entry(nonfixed_block_id1, nonfixed_block_id2, -1 * net_fanouts.get_entry(block_id1, col));
						//P.add_to_entry(nonfixed_block_id2, nonfixed_block_id1, -1 * net_fanouts.get_entry(block_id1, col));
					}
					else //block2 is fixed
					{
						//Ci = sum (Wiz * Xz), all fixed z
						C[nonfixed_block_id1] += net_fanouts.get_entry(block_id1, col) * block_fixed_id_to_x[fixed_block_id2];
						D[nonfixed_block_id1] += net_fanouts.get_entry(block_id1, col) * block_fixed_id_to_y[fixed_block_id2];
					}
				}
				if (nonfixed_block_id2 != -1) //nonfixed
				{
					//Qii = sum Wik, all k
					Q.add_to_entry(nonfixed_block_id2, nonfixed_block_id2, net_fanouts.get_entry(block_id2, col));
					//P.add_to_entry(nonfixed_block_id2, nonfixed_block_id2, net_fanouts.get_entry(block_id2, col));

					if (nonfixed_block_id1 == -1) //block1 is fixed
					{
						//Ci = sum (Wiz * Xz), all fixed z
						C[nonfixed_block_id2] += net_fanouts.get_entry(block_id2, col) * block_fixed_id_to_x[fixed_block_id1];
						D[nonfixed_block_id2] += net_fanouts.get_entry(block_id2, col) * block_fixed_id_to_y[fixed_block_id1];
					}
				}
			}
		}
		/*
		rows = net_fanouts.get_row_indicies(col, &num_rows);
		for (irow1 = 0; irow1 < num_rows; ++ irow1)
		{
			block_id1 = rows[irow1];
			nonfixed_block_id1 = block_id_to_nonfixed_id[block_id1];
			fixed_block_id1 = block_id_to_fixed_id[block_id1];
			for (irow2 = irow1 + 1; irow2 < num_rows; ++ irow2)
			{
				MYASSERT (rows[irow1] != rows[irow2]);
				block_id2 = rows[irow2];
				nonfixed_block_id2 = block_id_to_nonfixed_id[block_id2];
				fixed_block_id2 = block_id_to_fixed_id[block_id2];
				
				if (nonfixed_block_id1 != -1) //nonfixed
				{
					//Qii = sum Wik, all k
					Q.add_to_entry(nonfixed_block_id1, nonfixed_block_id1, net_fanouts.get_entry(block_id1, col));
					//P.add_to_entry(nonfixed_block_id1, nonfixed_block_id1, net_fanouts.get_entry(block_id1, col));

					if (nonfixed_block_id2 != -1) //also nonfixed
					{
						//Qij = Qji = -Wij
						Q.add_to_entry(nonfixed_block_id1, nonfixed_block_id2, -1 * net_fanouts.get_entry(block_id1, col));
						Q.add_to_entry(nonfixed_block_id2, nonfixed_block_id1, -1 * net_fanouts.get_entry(block_id1, col));
						//P.add_to_entry(nonfixed_block_id1, nonfixed_block_id2, -1 * net_fanouts.get_entry(block_id1, col));
						//P.add_to_entry(nonfixed_block_id2, nonfixed_block_id1, -1 * net_fanouts.get_entry(block_id1, col));
					}
					else //block2 is fixed
					{
						//Ci = sum (Wiz * Xz), all fixed z
						C[nonfixed_block_id1] += net_fanouts.get_entry(block_id1, col) * block_fixed_id_to_x[fixed_block_id2];
						D[nonfixed_block_id1] += net_fanouts.get_entry(block_id1, col) * block_fixed_id_to_y[fixed_block_id2];
					}
				}
				if (nonfixed_block_id2 != -1) //nonfixed
				{
					//Qii = sum Wik, all k
					Q.add_to_entry(nonfixed_block_id2, nonfixed_block_id2, net_fanouts.get_entry(block_id2, col));
					//P.add_to_entry(nonfixed_block_id2, nonfixed_block_id2, net_fanouts.get_entry(block_id2, col));

					if (nonfixed_block_id1 == -1) //block1 is fixed
					{
						//Ci = sum (Wiz * Xz), all fixed z
						C[nonfixed_block_id2] += net_fanouts.get_entry(block_id2, col) * block_fixed_id_to_x[fixed_block_id1];
						D[nonfixed_block_id2] += net_fanouts.get_entry(block_id2, col) * block_fixed_id_to_y[fixed_block_id1];
					}
				}
			}
		}
		*/
	}

	Q.debug_print("Q");
	
	/*
	debug_print_array(Q.m_ap.size(), vector_to_array(Q.m_ap), "%d ", "Ap is");
	debug_print_array(Q.m_rows.size(), vector_to_array(Q.m_rows), "%d ", "Ai is");
	debug_print_array(Q.m_vals.size(), vector_to_array(Q.m_vals), "%f ", "Ax is");
	*/
	debug_print_array(block_nonfixed_id_to_id.size(), C, "%f ", "C is");
	debug_print_array(block_nonfixed_id_to_id.size(), D, "%f ", "D is");


#ifndef MYUMF
	//solve for X
	int j;
	for (j = 0; j < block_nonfixed_id_to_id.size(); ++ j)
	{
		block_nonfixed_id_to_x[j] = rand() * CHIP_DIM / RAND_MAX;
		block_nonfixed_id_to_y[j] = rand() * CHIP_DIM / RAND_MAX;
	}
#endif

	
#ifdef MYUMF
   /*
	int n = 5 ;
   int Ap [ ] = {0, 2, 5, 9, 10, 12} ;
   int Ai [ ] = { 0, 1, 0, 2, 4, 1, 2, 3, 4, 2, 1, 4} ;
   double Ax [ ] = {2., 3., 3., -1., 4., 4., -3., 1., 2., 2., 6., 1.} ;
   double b [ ] = {8., 45., -3., 3., 19.} ;
   double x [5] ;

	double *null = (double *) NULL ;
	int i ;
   */

   /*void *Symbolic, *Numeric ;
   umfpack_di_symbolic (block_nonfixed_id_to_id.size(), block_nonfixed_id_to_id.size(), vector_to_array(Q.m_ap), vector_to_array(Q.m_rows), vector_to_array(Q.m_vals), &Symbolic, NULL, NULL) ;
	umfpack_di_numeric (vector_to_array(Q.m_ap), vector_to_array(Q.m_rows), vector_to_array(Q.m_vals), Symbolic, &Numeric, NULL, NULL) ;
	umfpack_di_free_symbolic (&Symbolic) ;
	umfpack_di_solve (UMFPACK_A, vector_to_array(Q.m_ap), vector_to_array(Q.m_rows), vector_to_array(Q.m_vals), block_nonfixed_id_to_x, C, Numeric, NULL, NULL) ;
   umfpack_di_solve (UMFPACK_A, vector_to_array(Q.m_ap), vector_to_array(Q.m_rows), vector_to_array(Q.m_vals), block_nonfixed_id_to_y, D, Numeric, NULL, NULL) ;
	umfpack_di_free_numeric (&Numeric) ;*/

	if (!use_cuda)
	{
		solve_serial(Q, block_nonfixed_id_to_x, C, block_nonfixed_id_to_y, D, block_nonfixed_id_to_id.size());
	}
	else
	{
		solve_cuda(Q, block_nonfixed_id_to_x, C, block_nonfixed_id_to_y, D, block_nonfixed_id_to_id.size());
	}

	debug_print_array(block_nonfixed_id_to_id.size(), block_nonfixed_id_to_x, "%f", "X is");
	debug_print_array(block_nonfixed_id_to_id.size(), block_nonfixed_id_to_y, "%f", "Y is ");

#endif



	//copy to actual x and y
	block_nonfixed_id_to_x_actual = new float[block_nonfixed_id_to_id.size()];
	block_nonfixed_id_to_y_actual = new float[block_nonfixed_id_to_id.size()];
	for (unsigned int i = 0; i < block_nonfixed_id_to_id.size(); ++ i)
	{
		block_nonfixed_id_to_x_actual[i] = block_nonfixed_id_to_x[i];
		block_nonfixed_id_to_y_actual[i] = block_nonfixed_id_to_y[i];
		//printf("block %d: (%f, %f)\n", block_id_to_user_name[i], block_nonfixed_id_to_x_actual[i], block_nonfixed_id_to_y_actual[i]);
	}
}


int main(int argc, char* argv[])
{
	if (argc != 4)
	{
		printf("Usage <filename> <use cuda> <display graphics>\n");
		exit(-1);
	}

	do_gordian = 'n';
	
	gordian_edge_weight = 0;

	run_fitting = 'n';

	fit_sort_by_fanout = 'n';
	
	num_random_io_swaps = 0;

	MYASSERT (argv[2][1] == '\0');
	MYASSERT (argv[2][0] == 'y' || argv[2][0] == 'n');
	use_cuda = (argv[2][0] == 'y');

	display_graphics = argv[3][0];
	MYASSERT (argv[3][1] == '\0');
	MYASSERT (display_graphics == 'n' || display_graphics == 'y');

	if (display_graphics != 'n')
	{
		/* initialize display with WHITE background, and define a clean_up function */
		init_graphics("Analytic Placer", WHITE, clean_up);

		/* still picture drawing allows user to zoom, etc. */
		init_world (0.,0.,WORLD_WIDTH,WORLD_HEIGHT);
	}

	
	
	read_netlist(argv[1]);

	randomize_ios();

	do_analytic_placement();
   
	float wl = 0.0;
	//wl = (float)compute_wire_length_hp();
	//sprintf(buffer, "After AP only - HP W/L = %f", wl);
	//my_update_message(buffer);
	my_event_loop(button_press, mouse_move, key_press, drawscreen);



	if (do_gordian == 'y')
	{
		do_gordian_overlap();

		wl = (float)compute_wire_length_hp();
		sprintf(buffer, "After Gordian Overlap removal - HP W/L = %f", wl);
		my_update_message(buffer);
		my_event_loop(button_press, mouse_move, key_press, drawscreen);
	}
	
	if (run_fitting == 'y')
	{
		run_fitter();

		wl = (float)compute_wire_length_hp();
		sprintf(buffer, "After fitter - HP W/L = %f", wl);
		my_update_message(buffer);
		my_event_loop(button_press, mouse_move, key_press, drawscreen);
	}

	

}

double compute_wire_length_hp()
{
	double wire_length = 0;
	//int col, *rows, num_rows, irow;
	//int block_id;
	double x,y;
	for (unsigned int col = 0; col < net_fanouts.get_num_cols(); ++ col)
	{
		double x_min = INT_MAX, y_min = INT_MAX, x_max = INT_MIN, y_max = INT_MIN;
		for (unsigned int irow = 0; irow < net_fanouts.get_num_rows(); ++ irow)
		{
			if (net_fanouts.get_entry(irow, col) == 0)
				continue;

			unsigned int block_id = irow;
			if (block_id_to_nonfixed_id[block_id] != -1) //non fixed
			{
				x = block_nonfixed_id_to_x_actual[block_id_to_nonfixed_id[block_id]];
				y = block_nonfixed_id_to_y_actual[block_id_to_nonfixed_id[block_id]];
			}
			else //fixed
			{
				x = block_fixed_id_to_x[block_id_to_fixed_id[block_id]];
				y = block_fixed_id_to_y[block_id_to_fixed_id[block_id]];
			}
			x_min = min(x, x_min);
			x_max = max(x, x_max);
			y_min = min(y, y_min);
			y_max = max(y, y_max);
		}
		MYASSERT(x_max >= x_min);
		MYASSERT(y_max >= y_min);
		wire_length += x_max - x_min + y_max - y_min;
	}
	return wire_length;
}

void read_netlist(char *filename)
{
	
	unsigned int block_num, net_num, row, col;

	FILE *in = fopen(filename, "r");

	if (in == NULL)
	{
		printf("Could not open file %s for reading\n", filename);
		exit(-1);
	}

	while (fscanf(in, "%d", &block_num) == 1)
	{
		if (block_num == -1)
		{
			break; //end of section
		}
		
		MYASSERT(block_num > 0);

		if (block_user_name_to_id.size() < block_num + 1)
		{
			block_user_name_to_id.resize(block_num + 1, -1);
		}
		MYASSERT(block_user_name_to_id[block_num] == -1);
		block_user_name_to_id[block_num] = block_id_to_user_name.size();
		block_id_to_user_name.push_back(block_num);

		while (fscanf(in, "%d", &net_num) == 1)
		{
			if (net_num == -1)
			{
				break; //end of line
			}
			MYASSERT(net_num > 0);

			if (net_user_name_to_id.size() < net_num + 1)
			{
				net_user_name_to_id.resize(net_num + 1, -1);
			}
			//new net encountered
			if (net_user_name_to_id[net_num] == -1)
			{
				net_user_name_to_id[net_num] = net_id_to_user_name.size();
				net_id_to_user_name.push_back(net_num);
				net_fanout_count.push_back(0);
			}

			++ net_fanout_count[net_user_name_to_id[net_num]];
		}
		if (net_num != -1)
		{
			MYDIE("COULD NOT READ IN net_num or line not terminated with -1");
		}
	}
	if (block_num != -1)
	{
		MYDIE("COULD NOT READ IN block_num or section not terminated with -1");
	}

	block_id_to_fixed_id.resize(block_id_to_user_name.size(), -1);

	while (fscanf(in, "%d %d %d", &block_num, &col, &row) == 3)
	{
		MYASSERT(row <= CHIP_DIM);
		MYASSERT(col <= CHIP_DIM);
		block_id_to_fixed_id[block_user_name_to_id[block_num]] = block_fixed_id_to_id.size();
		block_fixed_id_to_id.push_back(block_user_name_to_id[block_num]);
		block_fixed_id_to_x.push_back(col);
		block_fixed_id_to_y.push_back(row);
	}

	MYASSERT(block_num == -1);

	//now get a mapping of block ids to nonfixed ids
	block_id_to_nonfixed_id.resize(block_id_to_user_name.size(), -1);
	for (unsigned int i = 0; i < block_id_to_user_name.size(); ++ i)
	{
		if (block_id_to_fixed_id[i] == -1) //nonfixed block
		{
			block_id_to_nonfixed_id[i] = block_nonfixed_id_to_id.size();
			block_nonfixed_id_to_id.push_back(i);
		}
	}
	
	fclose (in);

	//store fanout of each net
	//net_fanouts.init(block_id_to_user_name.size(), net_id_to_user_name.size(), vector_to_array(net_fanout_count));
	net_fanouts.init(block_id_to_user_name.size(), net_id_to_user_name.size());

	//now go through placed block section
	//now that we know how many blocks there are, set up full weight matrix
	//in = fseek(in, 0, 
	in = fopen(filename, "r");

	while (fscanf(in, "%d", &block_num) == 1)
	{
		if (block_num == -1)
		{
			break; //end of section
		}
		while (fscanf(in, "%d", &net_num) == 1)
		{
			if (net_num == -1)
			{
				break; //end of line
			}
			//add to net fanout matrix
			net_fanouts.add_to_entry(block_user_name_to_id[block_num], net_user_name_to_id[net_num], 2.0/net_fanout_count[net_user_name_to_id[net_num]]);
//			net_fanouts.debug_print(block_id_to_user_name.size());
//			printf("\n\n\n");

		}
	}

	net_fanouts.debug_print("Net fanouts");

}

void randomize_ios()
{
	//swap placements of fixed blocks
	int num_fixed_blocks = block_fixed_id_to_id.size();
	int index1, index2, temp;
	int iswap;
	for (iswap = 0; iswap < num_random_io_swaps; ++ iswap)
	{
		index1 = rand() % num_fixed_blocks;
		index2 = rand() % num_fixed_blocks;

		temp = block_fixed_id_to_x[index1];
		block_fixed_id_to_x[index1] = block_fixed_id_to_x[index2];
		block_fixed_id_to_x[index2] = temp;

		temp = block_fixed_id_to_y[index1];
		block_fixed_id_to_y[index1] = block_fixed_id_to_y[index2];
		block_fixed_id_to_y[index2] = temp;
	}
}

//for linux ambiguity in abs function
double myabs(double x)
{
   if (x >= 0)
   {
      return x;
   }
   else
   {
      return -x;
   }
}

double calc_fit_error()
{
	double fit_error = 0.0f;
	int i;
	int num_nonfixed_blocks = block_nonfixed_id_to_id.size();
	for (i = 0; i < num_nonfixed_blocks; ++ i)
	{
		fit_error += 
			myabs(block_nonfixed_id_to_x[i] - block_nonfixed_id_to_x_actual[i]) +
			myabs(block_nonfixed_id_to_y[i] - block_nonfixed_id_to_y_actual[i]);
	}
	return fit_error;
}

int *block_nonfixed_total_fanout_count; //includes fanout to fixed blocks and multiple edges
int sort_by_decreasing_block_fanout(int x, int y)
{
	return block_nonfixed_total_fanout_count[y] < block_nonfixed_total_fanout_count[x];
}

void run_fitter()
{
	//need to keep track of every slot on the chip
	//if not too many blocks, can use set instead
	bool is_slot_used[CHIP_DIM + 1][CHIP_DIM + 1];
	int row, col;
	for (row = 0; row <= CHIP_DIM; ++ row)
	{
		for (col = 0; col <= CHIP_DIM; ++ col)
		{
			is_slot_used[row][col] = false;
		}
	}

	//mark fixed blocks spots
	int num_fixed_blocks = block_fixed_id_to_id.size();
	int i;
	for (i = 0; i < num_fixed_blocks; ++ i)
	{
		//make sure fixed blocks don't overlap
		//MYASSERT(is_slot_used[block_fixed_id_to_x[i]][block_fixed_id_to_y[i]] == false);
		is_slot_used[block_fixed_id_to_x[i]][block_fixed_id_to_y[i]] = true;
	}
	
	int num_nonfixed_blocks = block_nonfixed_id_to_id.size();

	//sort edges by fanout, ones with more fanout we want to give priority in fitting
	block_nonfixed_total_fanout_count = new int[num_nonfixed_blocks];
	for (i = 0; i < num_nonfixed_blocks; ++ i)
	{
		block_nonfixed_total_fanout_count[i] = 0;
	}

	//int *rows, num_rows, irow, block_id
	vector<int> connected_blocks;
	for (unsigned int col = 0; col < net_fanouts.get_num_cols(); ++ col)
	{
		connected_blocks.clear();
		unsigned int total_fanout_count = 0;
		for (unsigned int irow = 0; irow < net_fanouts.get_num_rows(); ++ irow)
		{
			if (net_fanouts.get_entry(irow, col) == 0)
				continue;
			
			unsigned block_id = irow;
			if (block_id_to_nonfixed_id[block_id] == -1)
			{
				continue; //is a fixed block - don't care
			}
			connected_blocks.push_back(block_id);
		}
		//this net connects this block to this many others
		for (unsigned int c = 0; c < connected_blocks.size(); c++)
		{
			block_nonfixed_total_fanout_count[block_id_to_nonfixed_id[connected_blocks[c]]] += connected_blocks.size() - 1;
		}
	}

	int *ordering = new int[num_nonfixed_blocks];
	for (i = 0; i < num_nonfixed_blocks; ++ i)
	{
		ordering[i] = i;
	}

	if (fit_sort_by_fanout != 'n')
	{
		sort(&ordering[0], &ordering[num_nonfixed_blocks], sort_by_decreasing_block_fanout);
	}

	if (fit_sort_by_fanout == 'i')
	{
		reverse(&ordering[0], &ordering[num_nonfixed_blocks]);
	}
	
	delete[] block_nonfixed_total_fanout_count;

	//for each block, try to place at closest spot by spiralling outward

	//simplest way to spiral - hard code vector offsets :)
	//try going right, left, up, down, up-right, down-right, up-left, down-left
	const int NUM_SUB_MOVES = 8;
	int x_offset[NUM_SUB_MOVES]		= {1,	0,	0,	0,	1,	0,	0,	0};
	int x_multiply[NUM_SUB_MOVES]	= {0,	-1,	0,	0,	0,	1,	-1,	1};
	int y_offset[NUM_SUB_MOVES]		= {0,	0,	1,	0,	1,	0,	0,	0};
	int y_multiply[NUM_SUB_MOVES]	= {0,	0,	0,	-1,	0,	-1,	-1,	-1};

	int x,y, move_count, distance, sub_move_count;
	for (i = 0; i < num_nonfixed_blocks; ++ i)
	{
		//round to closest
		x = (int)floor(block_nonfixed_id_to_x[ordering[i]] + 0.5f);
		y = (int)floor(block_nonfixed_id_to_y[ordering[i]] + 0.5f);

		MYASSERT(0 <= x && x <= CHIP_DIM);
		MYASSERT(0 <= y && y <= CHIP_DIM);

		int dx = 0, dy = 0;
		move_count = 0;
		while ( x + dx < 0 || x + dx > CHIP_DIM || //not valid slot - go to next one
				y + dy < 0 || y + dy > CHIP_DIM ||
				is_slot_used[x + dx][y + dy]) //valid slot but already full
		{
			distance = 1 + move_count / NUM_SUB_MOVES;
			sub_move_count = move_count % NUM_SUB_MOVES; //to index move array

			dx = x_multiply[sub_move_count] * dx + x_offset[sub_move_count] * distance;
			dy = y_multiply[sub_move_count] * dy + y_offset[sub_move_count] * distance;

			++ move_count;

			//watch out for infinite loop
			if (distance > CHIP_DIM)
			{
				MYDIE("Could not fit design - there are more blocks than slots on chip");
			}
		}

		MYASSERT(!is_slot_used[x + dx][y + dy]);
		is_slot_used[x + dx][y + dy] = true;
		block_nonfixed_id_to_x_actual[ordering[i]] = x + dx;
		block_nonfixed_id_to_y_actual[ordering[i]] = y + dy;
	}

	printf("Fit error is %f\n", calc_fit_error());
}
