#include "sparse_matrix.h"

sparse_matrix::sparse_matrix()
{
	m_rows = 0;
	m_columns = 0;
}
sparse_matrix::sparse_matrix(const sparse_matrix & rhs)
{
	m_matrix.resize(rhs.m_rows, NULL);
	for (unsigned int c = 0; c < m_matrix.size(); c++)
	{
		m_matrix[c] = new vector<float>(rhs.m_columns, 0);
		(*m_matrix[c]) = (*rhs.m_matrix[c]);
	}
	m_rows = rhs.m_rows;
	m_columns = rhs.m_columns;
}
sparse_matrix::~sparse_matrix()
{
	for (unsigned int c = 0; c < m_matrix.size(); c++)
	{
		delete m_matrix[c];
	}
}
void sparse_matrix::init(unsigned int rows, unsigned int columns)
{
	m_matrix.resize(rows, NULL);
	for (unsigned int c = 0; c < m_matrix.size(); c++)
	{
		m_matrix[c] = new vector<float>(columns, 0);
	}
	m_rows = rows;
	m_columns = columns;
}
void sparse_matrix::add_to_entry_expand(unsigned int row, unsigned int col, float value)
{
	MYASSERT(col >= 0 && col < m_columns);
	if (row >= m_rows)
	{
		//Add a new row
		m_matrix.resize(row+1);
		for (unsigned int c = m_rows-1; c < row+1; c++)
		{
			m_matrix[c] = new vector<float>(m_columns, 0);
		}
	}
	(*m_matrix[row])[col] += value;
}

void sparse_matrix::add_to_entry(unsigned int row, unsigned int col, float value)
{
	MYASSERT(col >= 0 && col < m_columns && row >= 0 && row < m_rows);
	(*m_matrix[row])[col] += value;
}
float sparse_matrix::get_entry(unsigned int row, unsigned int col)
{
	MYASSERT(col >= 0 && col < m_columns && row >= 0 && row < m_rows);
	return (*m_matrix[row])[col];
}
unsigned int sparse_matrix::get_num_cols() {return m_columns;}
unsigned int sparse_matrix::get_num_rows() {return m_rows;}
void sparse_matrix::swap(unsigned int row1, unsigned int row2)
{
	MYASSERT(row1 >= 0 && row1 < m_rows && row2 >= 0 && row2 < m_rows);
	m_matrix[row1]->swap(*m_matrix[row2]);
	//vector <float> * p = m_matrix[row1];
	//m_matrix[row1] = m_matrix[row2];
	//m_matrix[row2] = p;
}
void sparse_matrix::set_entry(unsigned int row, unsigned int col, float value)
{
	MYASSERT(row >= 0 && row < m_rows && col >= 0 && col < m_columns);
	(*m_matrix[row])[col] = value;
}
void sparse_matrix::debug_print(const char * description)
{
#ifdef MYDEBUG
	printf("\n\n"); printf(description); printf("\n");
	for (unsigned int row = 0; row < m_rows; row++)
	{
		for (unsigned int col = 0; col < m_columns; col++)
		{
			printf("%f\t", (*m_matrix[row])[col]);
		}
		printf("\n");
	}
#endif
}
